import pkg_resources
from py3o.fusion.log import logging
from twisted.web.template import Element, renderer
from twisted.web.resource import Resource
from twisted.web.template import flatten
from twisted.web.server import NOT_DONE_YET

from py3o.fusion.template import tloader


class IndexElement(Element):

    cssurls = [
        "/static/bootstrap/css/bootstrap.min.css",
        "/static/css/cover.css"
    ]

    loader = tloader('index.xml')
    header_loader = tloader('header.xml')
    scripts_loader = tloader('scripts.xml')

    @renderer
    def csslinks(self, request, tag):
        for cssurl in self.cssurls:
            yield tag.clone().fillSlots(cssurl=cssurl)

    @renderer
    def scripts(self, request, tag):
        return self.scripts_loader.load()

    @renderer
    def header(self, request, tag):
        return self.header_loader.load()

    @renderer
    def title(self, request, tag):
        return tag('py3o.fusion server')

    @renderer
    def version(self, request, tag):
        # return our own version so it can be displayed
        return tag(pkg_resources.get_distribution("py3o.fusion").version)


class RootPage(Resource):

    def flattened(self, output, request):
        # output should be None and ignored because flatten is as streamed data
        # to the connection as soon as it could
        # request.write('<!-- End of flow -->')
        request.finish()

    def render_GET(self, request):
        logging.info("GET request from {}".format(request.getClientIP()))
        request.write('<!DOCTYPE html>\n')
        d = flatten(request, IndexElement(), request.write)
        d.addCallback(self.flattened, request)
        return NOT_DONE_YET



from twisted.web.template import Element, renderer
from twisted.web.resource import Resource
from twisted.web.template import flatten
from twisted.web.template import XMLString
from twisted.web.server import NOT_DONE_YET
import pkg_resources
from pygments import highlight
from pygments.lexers import PythonLexer
from pygments.formatters import HtmlFormatter

from py3o.fusion.template import tloader


class FeaturesElement(Element):

    cssurls = [
        "/static/bootstrap/css/bootstrap.min.css",
        "/static/css/pygments.css",
    ]

    loader = tloader('features.xml')
    header_loader = tloader('header.xml')
    scripts_loader = tloader('scripts.xml')

    odt2pdf_s = open(
        pkg_resources.resource_filename(
            'py3o.fusion',
            'static/examples/odt2pdf.py',
        ),
        'rb'
    ).read()

    fusion2pdf_s = open(
        pkg_resources.resource_filename(
            'py3o.fusion',
            'static/examples/fusion2pdf.py',
        ),
        'rb'
    ).read()

    @renderer
    def odt2pdf(self, request, tag):
        hcode = highlight(
            self.odt2pdf_s,
            PythonLexer(),
            HtmlFormatter(encoding='utf-8'),
        )
        return XMLString(hcode).load()

    @renderer
    def fusion2pdf(self, request, tag):
        hcode = highlight(
            self.fusion2pdf_s,
            PythonLexer(),
            HtmlFormatter(encoding='utf-8'),
        )
        return XMLString(hcode).load()

    @renderer
    def csslinks(self, request, tag):
        for cssurl in self.cssurls:
            yield tag.clone().fillSlots(cssurl=cssurl)

    @renderer
    def scripts(self, request, tag):
        return self.scripts_loader.load()

    @renderer
    def header(self, request, tag):
        return self.header_loader.load()

    @renderer
    def title(self, request, tag):
        return tag('py3o.fusion features')


class FeaturesPage(Resource):

    def flattened(self, output, request):
        # output should be None and ignored because flatten is as streamed data
        # to the connection as soon as it could
        #request.write('<!-- End of flow -->')
        request.finish()

    def render_GET(self, request):
        request.write('<!DOCTYPE html>\n')
        d = flatten(request, FeaturesElement(), request.write)
        d.addCallback(self.flattened, request)
        return NOT_DONE_YET



from twisted.web.template import XMLFile
from twisted.python.filepath import FilePath
import pkg_resources


def tloader(basefile):
    return XMLFile(
        FilePath(
            pkg_resources.resource_filename(
                'py3o.fusion',
                'templates/%s' % basefile
            )
        )
    )



# you'll need to install requests to make this example work
# pip install --upgrade requests
# should do the trick
import requests
import json

# point the client to your own py3o.fusion server
url = 'http://localhost:8765/form'

# target formats you want... can be odt, pdf, doc, docx
targetformats = ["odt", "pdf", "doc", "docx"]


class MyEncoder1(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Item):
            obj = obj._asdict()
        else:
            obj = super(MyEncoder1, self).default(obj)

        return obj


class Item(object):
    def _asdict(self):
        return self.__dict__


items = list()

item1 = Item()
item1.val1 = 'Item1 Value1'
item1.val2 = 'Item1 Value2'
item1.val3 = 'Item1 Value3'
item1.Currency = 'EUR'
item1.Amount = '12345.35'
item1.InvoiceRef = '#1234'
items.append(item1)

for i in xrange(1000):
    item = Item()
    item.val1 = 'Item%s Value1' % i
    item.val2 = 'Item%s Value2' % i
    item.val3 = 'Item%s Value3' % i
    item.Currency = 'EUR'
    item.Amount = '6666.77'
    item.InvoiceRef = 'Reference #%04d' % i
    items.append(item)

document = Item()
document.total = '9999999999999.999'

data = dict(items=items, document=document)

data_s = json.dumps(data, cls=MyEncoder1)

for targetformat in targetformats:
    # open the files you need
    files = {
        'tmpl_file': open('templates/py3o_example_template.odt', 'rb'),
        'staticimage.img_logo': open('images/new_logo.png', 'rb'),
    }

    # fusion API needs those 3 keys
    fields = {
        "targetformat": targetformat,
        "datadict": data_s,
        "image_mapping": json.dumps({"staticimage.img_logo": "logo"}),
    }

    # and it needs to receive a POST with fields and files
    r = requests.post(url, data=fields, files=files)

    # TODO: handle error codes
    if r.status_code == 400:
        # server says we have a problem...
        # let's give the info back to our human friend
        print r.json()

    else:
        chunk_size = 1024
        # fusion server will stream an ODT file content back
        ext = targetformat.lower()
        with open('request_out.%s' % ext, 'wb') as fd:
            for chunk in r.iter_content(chunk_size):
                fd.write(chunk)

    files['tmpl_file'].close()
    files['staticimage.img_logo'].close()



from py3o.fusion.log import logging
import argparse

from twisted.web.server import Site
from twisted.web.static import File
from twisted.web.resource import Resource
from twisted.internet import reactor

import pkg_resources

from py3o.fusion.formpage import FormPage
from py3o.fusion.indexpage import RootPage
from py3o.fusion.featurespage import FeaturesPage


def cmd_line_server():
    logging.info("py3o.fusion server starting")
    argparser = argparse.ArgumentParser()

    argparser.add_argument(
        "-d",
        "--debug",
        dest="debug",
        action="store_true",
        default=False,
        help="Activate debug traceback on the client side",
    )

    argparser.add_argument(
        "-p", "--listenport",
        dest="listen_port",
        help="specify the PORT on which our service will listen",
        metavar="PORT",
        default=8765)

    argparser.add_argument(
        "-s", "--renderserver",
        dest="render_server",
        help="specify the hostname/ip of the render server",
        metavar="RENDERSERVER",
        default=None)

    argparser.add_argument(
        "-r", "--renderport",
        dest="render_port",
        help="specify the PORT on which the renderserver is available",
        metavar="RPORT",
        default=8994)

    args = argparser.parse_args()

    logging.info("listing on port: %s" % args.listen_port)
    start_server(args)


def start_server(options):

    reactor.suggestThreadPoolSize(30)

    root = Resource()
    root.putChild(
        "static",
        File(
            pkg_resources.resource_filename("py3o.fusion", "static")
        )
    )
    root.putChild("", RootPage())

    formpage = FormPage(
        options.render_server,
        int(options.render_port),
        clientdebug=options.debug
    )

    root.putChild("form", formpage)
    root.putChild("features", FeaturesPage())
    factory = Site(root)
    reactor.listenTCP(int(options.listen_port), factory)
    reactor.run(installSignalHandlers=1)



from twisted.web.static import File
import os


class FileAutoDestroy(File):
    """special twisted.web.static.File resource subclass
    to make it auto destroy the file once served
    """
    def open(self):
        return AutodestroyFile(self.path, 'r')


class AutodestroyFile(file):
    """a simple file like object that will autodestroy when you close() it...
    """

    def close(self):
        """we just make sure that the file will be destroyed when closed...
        """
        super(AutodestroyFile, self).close()
        os.unlink(self.name)



# import the wonderful requests lib
# if you need to intall it just try
# pip install --upgrade requests
import requests

# define where is your py3o.fusion endpoint
url = 'http://localhost:8765/form'

# open up the file and stuff it into a dictionary
# tmpl_file is a required field on the form. If you don't give
# it to the endpoint you'll receive an error back from it.
files = {
    'tmpl_file': open('templates/simple.odt', 'rb')
}

# then prepare the other fields of the form
# those 3 fields are also mandatory and failing to POST
# one of them will get you an error back from the server
#
# In this example you can see we leave the datadictionary
# and the image_mapping empty... This is because we won't
# send a template to the server but a simple plain
# old ODT file
fields = {
    "targetformat": 'pdf',
    "datadict": "{}",
    "image_mapping": "{}",
}

# finally POST our request on the endpoint
r = requests.post(url, data=fields, files=files)

# don't forget to close our orginal odt file
files['tmpl_file'].close()

# see if it is a success or a failure
# ATM the server only returns 400 errors... this may change
if r.status_code == 400:
    # server says we have an error...
    # this means it properly catched the error and nicely
    # gave us some info in a JSON answer...
    # let's use this fact
    print r.json()

else:
    # if we're here it is because we should receive a new
    # file back

    # let's stream the file back here...
    chunk_size = 1024

    # fusion server will stream an ODT file content back
    outname = 'request_out.%s' % 'pdf'
    with open(outname, 'wb') as fd:
        for chunk in r.iter_content(chunk_size):
            fd.write(chunk)

    # warn our dear user his file is ready
    print "Your file: %s is ready" % outname



import traceback
import socket
from py3o.fusion.log import logging

import os
import cgi
import json
import copy

from twisted.web.template import Element, renderer
from twisted.internet.threads import deferToThread
from twisted.web.template import flatten
from twisted.web.server import NOT_DONE_YET
from twisted.web.resource import Resource

from pyjon.utils import get_secure_filename
from py3o.template import Template
from py3o.renderclient import RenderClient
from py3o.formats import Formats


from py3o.fusion.autodestroy import FileAutoDestroy
from py3o.fusion.template import tloader

formats = Formats()
all_formats = formats.get_known_format_names()
# native formats mean formats we don't need to send
# to any renderserver to obtain the final file
native_formats = formats.get_known_format_names(nativeonly=True)
extended_formats = list(set(all_formats).difference(set(native_formats)))


class FormElement(Element):

    cssurls = [
        "/static/bootstrap/css/bootstrap.min.css",
    ]

    loader = tloader('form.xml')
    header_loader = tloader('header.xml')
    scripts_loader = tloader('scripts.xml')

    @renderer
    def csslinks(self, request, tag):
        for cssurl in self.cssurls:
            yield tag.clone().fillSlots(cssurl=cssurl)

    @renderer
    def scripts(self, request, tag):
        return self.scripts_loader.load()

    @renderer
    def header(self, request, tag):
        return self.header_loader.load()

    @renderer
    def title(self, request, tag):
        return tag('py3o.fusion test form')

    @renderer
    def formats(self, request, tag):
        return tag(", ".join(request.allowed_formats))


class FormPage(Resource):

    def __init__(self, render_server, render_port, clientdebug=False):
        Resource.__init__(self)

        self.allowed_formats = copy.copy(native_formats)
        self.render_server = render_server
        self.render_port = render_port
        self.clientdebug = clientdebug

        if render_server:
            # user specified a render server... we'll add extended formats to
            # allowed formats
            self.allowed_formats.extend(extended_formats)

    def flattened(self, output, request):
        # output should be None and ignored because flatten has streamed data
        # to the connection as soon as it could
        request.write('<!-- End of flow -->')
        request.finish()

    def render_GET(self, request):
        request.write('<!DOCTYPE html>\n')
        request.allowed_formats = self.allowed_formats
        d = flatten(request, FormElement(), request.write)
        d.addCallback(self.flattened, request)
        return NOT_DONE_YET

    def render_POST(self, request):
        logging.info("Request received launching thread")

        headers = request.getAllHeaders()
        form = cgi.FieldStorage(
            fp=request.content,
            headers=headers,
            environ={
                'REQUEST_METHOD': 'POST',
                'CONTENT_TYPE': headers['content-type'],
            }
        )

        if "skipfusion" in form:
            targetformat, error, reasons = self.get_target_format(form, [])
            fname, error, reasons = self.serialize_template(
                form, error, reasons
            )

            res = {
                'request': request,
                'outname': fname,
                'error': error,
                'reasons': reasons,
                'targetformat': targetformat,
                'outmime': formats.get_format(targetformat).mimetype
            }

            d = deferToThread(self.rendertarget, res)
            d.addCallback(self.rendered)
            d.addErrback(self.error)

        else:

            # defer blocking work to a thread and tell client to wait for it
            d = deferToThread(self.fusion, request, form)
            d.addCallback(self.fusionned)
            d.addErrback(self.error)

        request.notifyFinish().addErrback(self._cancelled, d)
        return NOT_DONE_YET

    def _cancelled(self, failure, deferred):
        deferred.cancel()
        logging.info("Client closed connection before end of request")

    def rendertarget(self, res):
        """this is a blocking method called into a thread...
        """
        newres = copy.copy(res)

        if not self.render_port or not self.render_server:
            newres['error'] = True
            newres['reasons'].append(
                "Renderserver missing config '%s:%s'" % (
                    self.render_server, self.render_port
                )
            )

        else:
            client = RenderClient(self.render_server, self.render_port)

            # TODO: when auth is actually implemented on the render server
            # we'll need to implement the command line options...
            client.login('toto', 'plouf')

            newres['outname'] = get_secure_filename()
            targetformat = res.get('targetformat')

            try:
                client.render(
                    res['outname'],
                    newres['outname'],
                    targetformat,
                )
                newres['outmime'] = formats.get_format(targetformat).mimetype

            # TODO: catch all other errors?
            except socket.error, e:
                tb_string = traceback.format_exc()
                newres['error'] = True
                newres['reasons'].append("Renderserver error: %s" % repr(e))
                if self.clientdebug:
                    newres['reasons'].append(
                        "Renderserver traceback: %s" % tb_string
                    )
                print tb_string
                logging.critical("Traceback in thread: %s" % tb_string)

            except Exception, e:
                newres['error'] = True
                newres['reasons'].append("Renderserver error: %s" % repr(e))

            finally:
                # remove intermediate file after rendering
                os.unlink(res['outname'])

        return newres

    def rendered(self, res):
        logging.info(
            "rendering thread finished for target %s" % res['targetformat']
        )

        if res.get('error', False):
            self.reply_error(res)

        else:
            self.reply_success(res)

    def fusionned(self, res):
        """will be called by twisted when thread with ODT processing
        will be finished
        """
        logging.info("fusion thread finished")

        if res.get('error', False):
            self.reply_error(res)

        else:
            if not res.get('targetformat') in native_formats:
                # we need to transform it to the asked format...
                d = deferToThread(self.rendertarget, res)
                d.addCallback(self.rendered)
                d.addErrback(self.error)
                return NOT_DONE_YET

            else:
                # we already have the desired format at hand
                # just return it
                self.reply_success(res)

    def reply_error(self, res):
        request = res.get('request')

        # bad request syntax...
        request.setResponseCode(400)
        # we got an error here... return something different from the ODT
        reasons = res.get('reasons', [])
        logging.info("Errors sent to client --> %s" % reasons)
        request.write(json.dumps({"error": True, "reasons": reasons}))
        request.finish()

    def reply_success(self, res):
        request = res.get('request')

        logging.info("returning autodestroy file resource")

        fileres = FileAutoDestroy(
            res['outname'],
            defaultType=res['outmime'],
        )
        # the FileAutoDestroy is a subclass of t.w.s.File and will
        # automatically call request.finish() for us at the end of
        # the file streaming
        return fileres.render_GET(request)

    def error(self, error):
        """if some client cancels the connection don't fret!
        """
        logging.error(error)

    def get_target_format(self, form, reasons):
        """helper function called by our thread
        """
        error = False
        if 'targetformat' in form and form["targetformat"].value:
            targetformat = form["targetformat"].value
        else:
            # if the user did not give give a target format...
            error = True
            reasons.append("targetformat must be specified")
            targetformat = "None given"

        # here we lower the requested format so that we compare with our
        # known formats which are always lower case. This allows old clients
        # to continue requesting upper case formats without knowing we changed
        if targetformat.lower() not in self.allowed_formats:
            error = True
            reasons.append(
                "target format: '{}' not supported by this server".format(
                    targetformat
                )
            )

        return targetformat, error, reasons

    def get_image_mapping(self, form, error, reasons):
        """helper function called by our thread
        """
        image_mapping = {}
        if "image_mapping" in form and form["image_mapping"].value:
            try:
                image_mapping = json.loads(
                    form["image_mapping"].value
                )
            except ValueError, e:
                error = True
                reasons.append("JSON parse error for image_mapping: %s" % e)

        else:
            error = True
            reasons.append("Image mapping must be given")

        return image_mapping, error, reasons

    def get_datadict(self, form, error, reasons):
        """helper function called by our thread
        """
        # if the user did not give a datadict... we assume it is empty
        datadict = {}
        if 'datadict' in form and form["datadict"].value:
            try:
                datadict = json.loads(
                    form["datadict"].value
                )
            except ValueError, e:
                error = True
                reasons.append("JSON parse error for datadict: %s" % e)

        return datadict, error, reasons

    def serialize_template(self, form, error, reasons):
        """helper function called by our thread
        """
        # grab file from POST into a local secure temp file
        fname = get_secure_filename()
        infile = open(fname, "wb")

        if 'tmpl_file' in form and form["tmpl_file"].value:
            infile.write(form["tmpl_file"].value)

        else:
            # caller did not provide a py3o.template... this is a clear cut
            # case where we must error & abort
            error = True
            reasons.append("No py3o template uploaded")

        infile.close()

        return fname, error, reasons

    def serialize_images(self, t, image_mapping, form, error, reasons):
        tempimages = []

        for imagename, imagevariable in image_mapping.iteritems():
            imname = get_secure_filename()
            infile = open(imname, "wb")
            if imagename in form:
                image_data = form[imagename].value

            else:
                image_data = None

            if not image_data:
                error = True
                reasons.append("Image %s was not uploaded" % imagename)
            else:
                infile.write(form[imagename].value)

            infile.close()
            tempimages.append(imname)

            if not error:
                t.set_image_path(imagevariable, imname)

        return tempimages, error, reasons

    def fusion(self, request, form):
        """WARNING, this method is called inside a thread...
        """
        error = False
        reasons = []

        targetformat, error, reasons = self.get_target_format(form, reasons)
        datadict, error, reasons = self.get_datadict(form, error, reasons)
        image_mapping, error, reasons = self.get_image_mapping(
            form, error, reasons
        )

        fname, error, reasons = self.serialize_template(form, error, reasons)

        # then fusion it with our py3o.template call
        outname = get_secure_filename()

        if error:
            os.unlink(fname)
            return {
                'request': request,
                'outname': outname,
                'error': error,
                'reasons': reasons,
                'targetformat': targetformat,
            }

        t = Template(fname, outname)

        tempimages, error, reasons = self.serialize_images(
            t, image_mapping, form, error, reasons
        )

        try:
            t.render(datadict)

        except ValueError, e:
            error = True
            # TODO: the repr is not perfect for image errors
            reasons.append(repr(e))

        except Exception, e:
            # unknow exceptions must be reported to client...
            error = True
            reasons.append(repr(e))

        # remove the input files after rendering
        os.unlink(fname)
        for tmpname in tempimages:
            os.unlink(tmpname)

        # then return the output result
        return {
            'request': request,
            'outname': outname,
            'error': error,
            'reasons': reasons,
            'targetformat': targetformat,
            'outmime': formats.get_format(targetformat).mimetype
        }



from setuptools import setup, find_packages

version = '0.8.2'

setup(
    name='py3o.fusion',
    version=version,
    description="A Fusion server that will transform your "
                "py3o.template into final LibreOffice documents",
    long_description=open("README.rst").read(),
    classifiers=[
        "Programming Language :: Python",
        "Topic :: Internet :: WWW/HTTP :: HTTP Servers",
        "Topic :: Text Processing :: General",
        "Development Status :: 4 - Beta",
        "Intended Audience :: Developers",
        "License :: OSI Approved :: BSD License",
        "Operating System :: OS Independent",
    ],
    keywords='LibreOffice OpenOffice templating PDF Fusion',
    author='Florent Aide',
    author_email='florent.aide@gmail.com',
    url='http://bitbucket.org/faide/py3o.fusion',
    license='BSD License',
    packages=find_packages(exclude=['ez_setup', 'examples', 'tests']),
    namespace_packages=['py3o'],
    include_package_data=True,
    zip_safe=True,
    install_requires=[
        'setuptools',
        'py3o.template >= 0.9.3',
        'py3o.renderclient >= 0.2',
        'py3o.formats >= 0.3',  # we need native flags from 0.3
        'twisted',
        'pygments',
    ],
    entry_points=dict(
        console_scripts=[
            'start-py3o-fusion = py3o.fusion.server:cmd_line_server',
        ],
    ),
)



#!/usr/bin/env python
# you'll need to install requests to make this example work
# pip install --upgrade requests
# should do the trick
import requests
import json

# point the client to your own py3o.fusion server
url = 'http://localhost:8765/form'

# target formats you want... can be odt, pdf, doc, docx
targetformats = ["pdf"]


class MyEncoder1(json.JSONEncoder):
    def default(self, obj):
        if isinstance(obj, Item):
            obj = obj._asdict()
        else:
            obj = super(MyEncoder1, self).default(obj)

        return obj


class Item(object):
    def _asdict(self):
        return self.__dict__


document = Item()
document.person_name = 'Aide'
document.person_surname = 'Florent'
document.person_company = "XCG Consulting"
document.person_url = "http://www.xcg-consulting.fr"

data = {
    "document": document
}

data_s = json.dumps(data, cls=MyEncoder1)

for targetformat in targetformats:
    # open the files you need
    files = {
        'tmpl_file': open('templates/fusion2pdf.odt', 'rb'),
    }

    # fusion API needs those 3 keys
    fields = {
        "targetformat": targetformat,
        "datadict": data_s,
        "image_mapping": "{}",
    }

    # and it needs to receive a POST with fields and files
    r = requests.post(url, data=fields, files=files)
    files['tmpl_file'].close()

    # TODO: handle error codes
    if r.status_code == 400:
        # server says we have a problem...
        # let's give the info back to our human friend
        print r.json()

    else:
        chunk_size = 1024
        # fusion server will stream an ODT file content back
        ext = targetformat.lower()
        with open('fusion2pdf.%s' % ext, 'wb') as fd:
            for chunk in r.iter_content(chunk_size):
                fd.write(chunk)
